/*
 * Copyright (c) Forge Development LLC and contributors
 * SPDX-License-Identifier: LGPL-2.1-only
 */

package net.minecraftforge.server;

import cpw.mods.modlauncher.InvalidLauncherSetupException;
import cpw.mods.modlauncher.Launcher;
import org.magmafoundation.magma.MagmaStart;
import org.magmafoundation.magma.betterui.BetterUI;
import org.magmafoundation.magma.updater.MagmaUpdater;
import org.yaml.snakeyaml.Yaml;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Paths;
import java.security.CodeSource;
import java.security.ProtectionDomain;
import java.util.Arrays;
import java.util.Optional;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

public class ServerMain {

    public static String[] oldArgs = new String[0];


    public static void main(String[] args) throws MalformedURLException, ClassNotFoundException, NoSuchMethodException, InvocationTargetException, InstantiationException, IllegalAccessException {
        oldArgs = args;
        System.setProperty("log4j2.formatMsgNoLookups", "true");

        try {
            if (Arrays.stream(args).anyMatch(s -> s.equalsIgnoreCase("-noui"))) {
                BetterUI.setEnabled(false);
                args = MagmaStart.remove(args, "-noui");
            }
            if (Arrays.stream(args).anyMatch(s -> s.equalsIgnoreCase("-nologo"))) {
                BetterUI.setEnableBigLogo(false);
                args = MagmaStart.remove(args, "-nologo");
            }
            if (Arrays.stream(args).anyMatch(s -> s.equalsIgnoreCase("-accepteula"))) {
                BetterUI.forceAcceptEULA(Paths.get("eula.txt"));
                args = MagmaStart.remove(args, "-accepteula");
            }
            if (Arrays.stream(args).anyMatch(s -> s.equalsIgnoreCase("-dau"))) {
                MagmaStart.enableUpdate = false;
                args = MagmaStart.remove(args, "-dau");
            }
            if (Arrays.stream(args).anyMatch(s -> s.equalsIgnoreCase("-nojline"))) {
                args = MagmaStart.remove(args, "-nojline"); //For some reason when passing -nojline to the console the whole thing crashes, remove this
            }
            MagmaStart.main(args);
            Class.forName("cpw.mods.modlauncher.Launcher", false, ClassLoader.getSystemClassLoader());
            Class.forName("net.minecraftforge.forgespi.Environment", false, ClassLoader.getSystemClassLoader());
        } catch (Exception cnfe) {
            System.err.println("There was an error loading the forge launcher. This is not a valid forge installation.");
            cnfe.printStackTrace();
            System.exit(-1);
        }

        final String launchArgs = Optional.ofNullable(ServerMain.class.getProtectionDomain()).
                map(ProtectionDomain::getCodeSource).
                map(CodeSource::getLocation).
                map(ServerMain::urlToManifest).
                map(Manifest::getMainAttributes).
                map(a -> a.getValue("ServerLaunchArgs")).
                orElseThrow(ServerMain::throwMissingManifest);
        String[] defaultargs = launchArgs.split(" ");
        String[] result = new String[args.length + defaultargs.length];
        System.arraycopy(defaultargs, 0, result, 0, defaultargs.length);
        System.arraycopy(args, 0, result, defaultargs.length, args.length);
        // separate class, so the exception can resolve

        new Runner().runLauncher(result);
    }

    private static class Runner {
        private void runLauncher(final String[] result) {
            try {
                Launcher.main(result);
            } catch (InvalidLauncherSetupException e) {
                System.err.println("The server is missing critical libraries and cannot load. Please run the installer to correct this");
                System.exit(1);
            }
        }
    }
    private static Manifest urlToManifest(URL url) {
        try {
            return new JarFile(new File(url.toURI())).getManifest();
        } catch (URISyntaxException | IOException e) {
            return null;
        }
    }

    private static RuntimeException throwMissingManifest() {
        System.err.println("This is not being run from a valid JAR file, essential data is missing.");
        return new RuntimeException("Missing the manifest");
    }
}
