package org.magmafoundation.magma.libs;

import me.tongfei.progressbar.ProgressBar;
import me.tongfei.progressbar.ProgressBarBuilder;
import me.tongfei.progressbar.ProgressBarStyle;
import org.magmafoundation.magma.MagmaConstants;
import org.magmafoundation.magma.utils.JarUtils;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.jar.Attributes;
import java.util.jar.JarFile;

public class MagmaInstaller {

    private static final String
            forgeVersion = MagmaConstants.FORGE_VERSION_FULL,
            mcpVersion = "20210115.111550";

    private static final String libPath = String.format("%s/%s/", JarUtils.getJarDir(), MagmaConstants.INSTALLER_LIBRARIES_FOLDER);

    private static final String forge = libPath + "net/minecraftforge/forge/";
    private static final String forgeBase = String.format("%s%s/forge-%s", forge, forgeVersion, forgeVersion);

    public static final File
            universalJar = new File(forgeBase + "-universal.jar"),
            serverJar = new File(forgeBase + "-server.jar");

    private static final File lzma = new File(libPath + "magma/install/data/server.lzma");

    private static final String otherStart = libPath + "net/minecraft/server/1.16.5-" + mcpVersion + "/server-1.16.5-" + mcpVersion;
    private static final File extra = new File(otherStart + "-extra.jar");

    private static final File slim = new File(otherStart + "-slim.jar");
    private static final File srg = new File(otherStart + "-srg.jar");

    private static final String mcpStart = libPath + "de/oceanlabs/mcp/mcp_config/1.16.5-" + mcpVersion + "/mcp_config-1.16.5-" + mcpVersion;
    private static final File mcpZip = new File(mcpStart + ".zip");
    private static final File mcpTxt = new File(mcpStart + "-mappings.txt");

    private static final PrintStream out = System.out;
    private static final PrintStream log;

    private static final File installerTools = new File(libPath + "net/minecraftforge/installertools/1.1.11/installertools-1.1.11-fatjar.jar");
    private static final File jarSpliter = new File(libPath + "net/minecraftforge/jarsplitter/1.1.2/jarsplitter-1.1.2-fatjar.jar");
    private static final File specialSource = new File(libPath + "net/md-5/SpecialSource/1.8.5/SpecialSource-1.8.5-shaded.jar");
    private static final File binaryPatches = new File(libPath + "net/minecraftforge/binarypatcher/1.0.12/binarypatcher-1.0.12-fatjar.jar");

    private static ClassLoader parentClassLoader;

    static {
        try {
            log = new PrintStream(new FileOutputStream("logs/installer.log"));
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    public static void install() throws Exception {
        ProgressBarBuilder builder = new ProgressBarBuilder()
                .setInitialMax(6)
                .setUpdateIntervalMillis(100)
                .setTaskName("Patching server...")
                .setStyle(ProgressBarStyle.ASCII);

        try (ProgressBar pb = builder.build()) {
            extractFromMagmaJar("data/server.lzma", lzma);
            pb.step();

            extractFromMagmaJar("data/forge-" + forgeVersion + "-universal.jar", universalJar);
            pb.step();

            if (mcpZip.exists()) {
                if (!mcpTxt.exists())
                    runInstallerTools();
            } else {
                System.err.println("MCP zip not found!");
                System.exit(-1);
            }
            pb.step();

            if (!slim.exists() || !extra.exists())
                runJarSplitter();
            pb.step();

            if (!srg.exists())
                runSpecialSource();
            pb.step();

            if (!serverJar.exists())
                runBinaryPatcher();
            pb.step();
        } catch (Exception e) {
            System.err.println("An error occurred while patching the server!");
            e.printStackTrace();
            System.exit(-1);
        }
        System.setOut(out);
        System.out.println("Patching complete! Starting server...");
        TimeUnit.SECONDS.sleep(1);
    }

    private static void runInstallerTools() throws Exception {
        System.setOut(log);
        System.out.println("Running installertools...");
        boolean success = launchService(installerTools, new String[]{
                "--task", "MCP_DATA",
                "--input", mcpZip.getAbsolutePath(),
                "--output", mcpTxt.getAbsolutePath(),
                "--key", "mappings"
        });
        System.setOut(out);
        if (!success) {
            System.err.println("InstallerTools failed!");
            System.exit(-1);
        }
    }

    private static void runJarSplitter() throws Exception {
        System.setOut(log);
        System.out.println("Running jarsplitter...");
        boolean success = launchService(jarSpliter, new String[]{
                "--input", libPath + "minecraft_server.1.16.5.jar",
                "--slim", slim.getAbsolutePath(),
                "--extra", extra.getAbsolutePath(),
                "--srg", mcpTxt.getAbsolutePath()
        });
        System.setOut(out);
        if (!success) {
            System.err.println("JarSplitter failed!");
            System.exit(-1);
        }
    }

    private static void runSpecialSource() throws Exception {
        System.setOut(log);
        System.out.println("Running specialsource...");
        boolean success = launchService(specialSource, new String[]{
                "--in-jar", slim.getAbsolutePath(),
                "--out-jar", srg.getAbsolutePath(),
                "--srg-in", mcpTxt.getAbsolutePath()
        });
        System.setOut(out);
        if (!success) {
            System.err.println("SpecialSource failed!");
            System.exit(-1);
        }
    }

    private static void runBinaryPatcher() throws Exception {
        System.setOut(log);
        System.out.println("Running binarypatcher...");
        boolean success = launchService(binaryPatches, new String[]{
                "--clean",
                srg.getAbsolutePath(),
                "--output",
                serverJar.getAbsolutePath(),
                "--apply",
                lzma.getAbsolutePath()
        });
        System.setOut(out);
        if (!success) {
            System.err.println("BinaryPatcher failed!");
            System.exit(-1);
        }
    }

    private static boolean launchService(File jar, String[] args) throws Exception {
        JarFile jarFile = new JarFile(jar);
        String mainClass = jarFile.getManifest().getMainAttributes().getValue(Attributes.Name.MAIN_CLASS);
        if (mainClass == null || mainClass.isEmpty())
            throw new RuntimeException("No main class found in " + jar.getAbsolutePath());

        List<URL> classpath = new ArrayList<>();
        classpath.add(jar.toURI().toURL());
        jarFile.close();

        ClassLoader loader = new URLClassLoader(classpath.toArray(new URL[classpath.size()]), getParentClassLoader());
        Thread currentThread = Thread.currentThread();
        ClassLoader threadClassLoader = currentThread.getContextClassLoader();
        currentThread.setContextClassLoader(loader);

        try {
            final Class<?> aClass = Class.forName(mainClass, true, loader);
            final Method method = aClass.getMethod("main", String[].class);
            method.invoke(null, (Object) args);
            return true;
        } catch (InvocationTargetException e) {
            final Throwable cause = e.getCause();
            cause.printStackTrace();
        } finally {
            currentThread.setContextClassLoader(threadClassLoader);
        }
        return false;
    }

    private static synchronized ClassLoader getParentClassLoader() {
        if (parentClassLoader == null) {
            if (!System.getProperty("java.version").startsWith("1.")) {
                try {
                    final Method getPlatform = ClassLoader.class.getDeclaredMethod("getPlatformClassLoader", new Class[0]);
                    MagmaInstaller.parentClassLoader = (ClassLoader)getPlatform.invoke(null, new Object[0]);
                } catch (NoSuchMethodException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
                    System.err.println("No platform classloader: " + System.getProperty("java.version"));
                }
            }
        }
        return parentClassLoader;
    }

    private static void extractFromMagmaJar(String fileName, File destination) throws IOException {
        destination.getParentFile().mkdirs();
        destination.createNewFile();

        try (InputStream in = MagmaInstaller.class.getClassLoader().getResourceAsStream(fileName)) {
            if (in == null) {
                System.err.println("Failed to find file: " + fileName + ", is the server corrupt?");
                System.exit(-1);
            }
            Files.copy(in, destination.toPath(), StandardCopyOption.REPLACE_EXISTING);
        }
    }

    public static File getExtras() {
        return extra;
    }
}
