package org.magmafoundation.magma.permission;

import com.google.common.base.Joiner;
import com.mojang.brigadier.ParseResults;
import com.mojang.brigadier.tree.CommandNode;
import net.minecraft.command.CommandSource;
import net.minecraft.command.Commands;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.dedicated.DedicatedServer;
import org.apache.commons.lang3.Validate;
import org.bukkit.Location;
import org.bukkit.command.*;
import org.bukkit.command.defaults.BukkitCommand;
import org.bukkit.craftbukkit.v1_16_R3.CraftServer;
import org.bukkit.craftbukkit.v1_16_R3.command.CraftBlockCommandSender;
import org.bukkit.craftbukkit.v1_16_R3.command.ProxiedNativeCommandSender;
import org.bukkit.craftbukkit.v1_16_R3.entity.CraftMinecartCommand;
import org.bukkit.craftbukkit.v1_16_R3.entity.CraftPlayer;
import org.bukkit.entity.Player;
import org.bukkit.entity.minecart.CommandMinecart;
import org.magmafoundation.magma.configuration.MagmaConfig;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class ForgeCommandWrapper extends BukkitCommand {

    private final Commands dispatcher;
    public final CommandNode<CommandSource> forgeCommand;

    public ForgeCommandWrapper(Commands dispatcher, CommandNode<CommandSource> forgeCommand) {
        super(forgeCommand.getName(), "A Forge modded command.", forgeCommand.getUsageText(), Collections.EMPTY_LIST);
        this.dispatcher = dispatcher;
        this.forgeCommand = forgeCommand;
        this.setPermission(getPermission(forgeCommand));
    }

    @Override
    public boolean execute(CommandSender sender, String commandLabel, String[] args) {
        if (!MagmaConfig.instance.forgeCommandsIgnoreBukkitPerms.getValues()) {
            //If we detect that the CraftPlayer permissible was injected into (ex. LuckPerms), we will use their permissions
            //instead of the Forge permissions defined in 'Commands.literal(...).requires(s -> s.hasPermission(..))'.
            //This will cause a problem though, as the Forge permissions will not be checked.
            //So you will have to manually set the Forge permissions by allowing ex. "forge.command.coolmodcommand" in LuckPerms.
            boolean permissibleInjected = false;
            if (sender instanceof CraftPlayer)
                permissibleInjected = ((CraftPlayer) sender).isPermissibleInjected();

            if ((permissibleInjected && !testPermission(sender)) ||
                    (!permissibleInjected && !forgeCommand.getRequirement().test(getListener(sender))))
                return true;
        } else {
            if (!forgeCommand.getRequirement().test(getListener(sender))) return true;
        }

        CommandSource icommandlistener = getListener(sender);
        dispatcher.setForge(true);
        dispatcher.handleCommand(icommandlistener, toDispatcher(args, getName()), toDispatcher(args, commandLabel), true);
        dispatcher.setForge(false);
        return true;
    }

    @Override
    public List<String> tabComplete(CommandSender sender, String alias, String[] args, Location location) throws IllegalArgumentException {
        org.apache.commons.lang3.Validate.notNull(sender, "Sender cannot be null");
        org.apache.commons.lang3.Validate.notNull(args, "Arguments cannot be null");
        Validate.notNull(alias, "Alias cannot be null");

        CommandSource icommandlistener = getListener(sender);
        ParseResults<CommandSource> parsed = dispatcher.getForgeDispatcher().parse(toDispatcher(args, getName()), icommandlistener);

        List<String> results = new ArrayList<>();
        dispatcher.getForgeDispatcher().getCompletionSuggestions(parsed).thenAccept((suggestions) -> {
            suggestions.getList().forEach((s) -> results.add(s.getText()));
        });

        return results;
    }

    public static CommandSource getListener(CommandSender sender) {
        if (sender instanceof Player) {
            return ((CraftPlayer) sender).getHandle().createCommandSourceStack();
        }
        if (sender instanceof BlockCommandSender) {
            return ((CraftBlockCommandSender) sender).getWrapper();
        }
        if (sender instanceof CommandMinecart) {
            return ((CraftMinecartCommand)sender).getHandle().getCommandBlock().createCommandSourceStack();
        }
        if (sender instanceof RemoteConsoleCommandSender) {
            return ((DedicatedServer) MinecraftServer.getServer()).rconConsoleSource.createCommandSourceStack();
        }
        if (sender instanceof ConsoleCommandSender) {
            return ((CraftServer) sender.getServer()).getServer().createCommandSourceStack();
        }
        if (sender instanceof ProxiedCommandSender) {
            return ((ProxiedNativeCommandSender) sender).getHandle();
        }

        throw new IllegalArgumentException("Cannot make " + sender + " a forge command listener");
    }

    public static String getPermission(CommandNode<CommandSource> forgeCommand) {
        return "forge.command." + ((forgeCommand.getRedirect() == null) ? forgeCommand.getName() : forgeCommand.getRedirect().getName());
    }

    private String toDispatcher(String[] args, String name) {
        return name + ((args.length > 0) ? " " + Joiner.on(' ').join(args) : "");
    }
}
