/*
 * Magma Server
 * Copyright (C) 2019-2022.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package org.magmafoundation.magma.util;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import jline.internal.Nullable;
import sun.misc.Unsafe;

/**
 * EnumHelper
 *
 * @author Hexeption admin@hexeption.co.uk
 * @since 08/06/2020 - 05:59 pm
 */
public class EnumHelper {

	@Nullable
	public static <T extends Enum<?>> T addEnum0(Class<T> enumType, String enumName, Class<?>[] pramsTypes, Object... paramValues) {
		return addEnum(enumType, enumName, pramsTypes, paramValues);
	}

	@Nullable
	public static <T extends Enum<?>> T addEnum(Class<T> enumType, String enumName, Class<?>[] paramTypes, Object[] paramValues) {
		if (Float.parseFloat(System.getProperty("java.class.version")) != 52.0) {
			return Java9.addEnum(enumType, enumName, paramTypes, paramValues);
		} else {
			return Java8.addEnum(enumType, enumName, paramTypes, paramValues);
		}
	}

	public static class Java8 {
		private static Object reflectionFactory = null;
		private static Method newConstructorAccessor = null;
		private static Method newInstance = null;
		private static Method newFieldAccessor = null;
		private static Method fieldAccessorSet = null;
		private static boolean isSetup = false;

		private static void setup() {
			if (isSetup) {
				return;
			}

			try {
				Method getReflectionFactory = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("getReflectionFactory");
				reflectionFactory = getReflectionFactory.invoke(null);
				newConstructorAccessor = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("newConstructorAccessor", Constructor.class);
				newInstance = Class.forName("sun.reflect.ConstructorAccessor").getDeclaredMethod("newInstance", Object[].class);
				newFieldAccessor = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("newFieldAccessor", Field.class, boolean.class);
				fieldAccessorSet = Class.forName("sun.reflect.FieldAccessor").getDeclaredMethod("set", Object.class, Object.class);
			} catch (Exception e) {
				e.printStackTrace();
			}

			isSetup = true;
		}

		/*
		 * Everything below this is found at the site below, and updated to be able to compile in Eclipse/Java 1.6+
		 * Also modified for use in decompiled code.
		 * Found at: http://niceideas.ch/roller2/badtrash/entry/java_create_enum_instances_dynamically
		 */
		private static Object getConstructorAccessor(Class<?> enumClass, Class<?>[] additionalParameterTypes) throws Exception {
			Class<?>[] parameterTypes = new Class[additionalParameterTypes.length + 2];
			parameterTypes[0] = String.class;
			parameterTypes[1] = int.class;
			System.arraycopy(additionalParameterTypes, 0, parameterTypes, 2, additionalParameterTypes.length);
			return newConstructorAccessor.invoke(reflectionFactory, enumClass.getDeclaredConstructor(parameterTypes));
		}

		private static <T extends Enum<?>> T makeEnum(Class<T> enumClass, String value, int ordinal, Class<?>[] additionalTypes, Object[] additionalValues) throws Exception {
			Object[] parms = new Object[additionalValues.length + 2];
			parms[0] = value;
			parms[1] = Integer.valueOf(ordinal);
			System.arraycopy(additionalValues, 0, parms, 2, additionalValues.length);
			return enumClass.cast(newInstance.invoke(getConstructorAccessor(enumClass, additionalTypes), new Object[] {parms}));
		}

		public static void setFailsafeFieldValue(Field field, Object target, Object value) throws Exception {
			field.setAccessible(true);
			Field modifiersField = Field.class.getDeclaredField("modifiers");
			modifiersField.setAccessible(true);
			modifiersField.setInt(field, field.getModifiers() & ~Modifier.FINAL);
			Object fieldAccessor = newFieldAccessor.invoke(reflectionFactory, field, false);
			fieldAccessorSet.invoke(fieldAccessor, target, value);
		}

		private static void blankField(Class<?> enumClass, String fieldName) throws Exception {
			for (Field field : Class.class.getDeclaredFields()) {
				if (field.getName().contains(fieldName)) {
					field.setAccessible(true);
					setFailsafeFieldValue(field, enumClass, null);
					break;
				}
			}
		}

		private static void cleanEnumCache(Class<?> enumClass) throws Exception {
			blankField(enumClass, "enumConstantDirectory");
			blankField(enumClass, "enumConstants");
		}

		public static <T extends Enum<?>> T addEnum(Class[][] map, Class<T> enumType, String enumName, Object... paramValues) {
			for (Class[] lookup : map) {
				if (lookup[0] == enumType) {
					Class<?>[] paramTypes = new Class<?>[lookup.length - 1];
					if (paramTypes.length > 0) {
						System.arraycopy(lookup, 1, paramTypes, 0, paramTypes.length);
					}
					return addEnum(enumType, enumName, paramTypes, paramValues);
				}
			}
			return null;
		}

		@SuppressWarnings("unchecked")
		public static <T extends Enum<?>> T addEnum(Class<T> enumType, String enumName, Class<?>[] paramTypes, Object[] paramValues) {
			if (!isSetup) {
				setup();
			}

			Field valuesField = null;
			Field[] fields = enumType.getDeclaredFields();
			int flags = Modifier.PRIVATE | Modifier.STATIC | Modifier.FINAL | 0x1000 /*SYNTHETIC*/;
			String valueType = String.format("[L%s;", enumType.getName().replace('.', '/'));

			for (Field field : fields) {
				if ((field.getModifiers() & flags) == flags &&
						field.getType().getName().replace('.', '/').equals(valueType)) //Apparently some JVMs return .'s and some don't..
				{
					valuesField = field;
					break;
				}
			}
			valuesField.setAccessible(true);

			try {
				T[] previousValues = (T[]) valuesField.get(enumType);
				List<T> values = new ArrayList<T>(Arrays.asList(previousValues));
				T newValue = (T) makeEnum(enumType, enumName, values.size(), paramTypes, paramValues);
				values.add(newValue);
				setFailsafeFieldValue(valuesField, null, values.toArray((T[]) Array.newInstance(enumType, 0)));
				cleanEnumCache(enumType);

				return newValue;
			} catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException(e.getMessage(), e);
			}
		}

		static {
			if (!isSetup) {
				setup();
			}
		}
	}

	public static class Java9 {
		private static MethodHandles.Lookup impl_lookup = null;
		private static boolean isSetup = false;
		public static sun.misc.Unsafe unsafe;

		private static void setup() {
			if (isSetup) {
				return;
			}

			try {
				Class<?> unsafeClass = Class.forName("sun.misc.Unsafe");
				Field theUnsafe = unsafeClass.getDeclaredField("theUnsafe");
				theUnsafe.setAccessible(true);
				unsafe = (Unsafe) theUnsafe.get(null);
				Field impl_lookupField = MethodHandles.Lookup.class.getDeclaredField("IMPL_LOOKUP");
				impl_lookup = (MethodHandles.Lookup) unsafe.getObject(unsafe.staticFieldBase(impl_lookupField), unsafe.staticFieldOffset(impl_lookupField));
			} catch (Exception e) {
				e.printStackTrace();
			}

			isSetup = true;
		}

		/*
		 * Everything below this is found at the site below, and updated to be able to compile in Eclipse/Java 1.6+
		 * Also modified for use in decompiled code.
		 * Found at: http://niceideas.ch/roller2/badtrash/entry/java_create_enum_instances_dynamically
		 */
		private static MethodHandle getConstructorAccessor(Class<?> enumClass, Class<?>[] additionalParameterTypes) throws Exception {
			Class<?>[] parameterTypes = new Class[additionalParameterTypes.length + 2];
			parameterTypes[0] = String.class;
			parameterTypes[1] = int.class;
			System.arraycopy(additionalParameterTypes, 0, parameterTypes, 2, additionalParameterTypes.length);
			return impl_lookup.findConstructor(enumClass, MethodType.methodType(void.class, parameterTypes));
		}

		private static <T extends Enum<?>> T makeEnum(Class<T> enumClass, String value, int ordinal, Class<?>[] additionalTypes, Object[] additionalValues) throws Throwable {
			Object[] parms = new Object[additionalValues.length + 2];
			parms[0] = value;
			parms[1] = Integer.valueOf(ordinal);
			System.arraycopy(additionalValues, 0, parms, 2, additionalValues.length);
			return enumClass.cast(getConstructorAccessor(enumClass, additionalTypes).invokeWithArguments(parms));
		}

		public static void setFailsafeFieldValue(Field field, Object target, Object value) throws Throwable {
			if (target != null) {
				impl_lookup.findSetter(field.getDeclaringClass(), field.getName(), field.getType()).invoke(target, value);
			} else {
				impl_lookup.findStaticSetter(field.getDeclaringClass(), field.getName(), field.getType()).invoke(value);
			}
		}

		private static void blankField(Class<?> enumClass, String fieldName) throws Throwable {
			for (Field field : Class.class.getDeclaredFields()) {
				if (field.getName().contains(fieldName)) {
					field.setAccessible(true);
					setFailsafeFieldValue(field, enumClass, null);
					break;
				}
			}
		}

		private static void cleanEnumCache(Class<?> enumClass) throws Throwable {
			blankField(enumClass, "enumConstantDirectory");
			blankField(enumClass, "enumConstants");
			blankField(enumClass, "enumVars");
		}

		public static <T extends Enum<?>> T addEnum(Class[][] map, Class<T> enumType, String enumName, Object... paramValues) {
			for (Class[] lookup : map) {
				if (lookup[0] == enumType) {
					Class<?>[] paramTypes = new Class<?>[lookup.length - 1];
					if (paramTypes.length > 0) {
						System.arraycopy(lookup, 1, paramTypes, 0, paramTypes.length);
					}
					return addEnum(enumType, enumName, paramTypes, paramValues);
				}
			}
			return null;
		}

		@SuppressWarnings("unchecked")
		public static <T extends Enum<?>> T addEnum(Class<T> enumType, String enumName, Class<?>[] paramTypes, Object[] paramValues) {
			if (!isSetup) {
				setup();
			}

			Field valuesField = null;
			Field[] fields = enumType.getDeclaredFields();
			int flags = Modifier.PRIVATE | Modifier.STATIC | Modifier.FINAL | 0x1000 /*SYNTHETIC*/;
			String valueType = String.format("[L%s;", enumType.getName().replace('.', '/'));

			for (Field field : fields) {
				if ((field.getModifiers() & flags) == flags &&
						field.getType().getName().replace('.', '/').equals(valueType)) //Apparently some JVMs return .'s and some don't..
				{
					valuesField = field;
					break;
				}
			}
			valuesField.setAccessible(true);

			try {
				T[] previousValues = (T[]) valuesField.get(enumType);
				List<T> values = new ArrayList<T>(Arrays.asList(previousValues));
				T newValue = (T) makeEnum(enumType, enumName, values.size(), paramTypes, paramValues);
				values.add(newValue);
				setFailsafeFieldValue(valuesField, null, values.toArray((T[]) Array.newInstance(enumType, 0)));
				cleanEnumCache(enumType);

				return newValue;
			} catch (Throwable throwable) {
				throwable.printStackTrace();
				throw new RuntimeException(throwable.getMessage(), throwable);
			}
		}

		static {
			if (!isSetup) {
				setup();
			}
		}


		public static void setField(Object obj, Object value, Field field) throws ReflectiveOperationException {
			if (obj == null) {
				setStaticField(field, value);
			} else {
				try {
					unsafe.putObject(obj, unsafe.objectFieldOffset(field), value);
				} catch (Exception e) {
					throw new ReflectiveOperationException(e);
				}
			}
		}

		public static void setStaticField(Field field, Object value) throws ReflectiveOperationException {
			try {
				unsafe.ensureClassInitialized(field.getDeclaringClass());
				unsafe.putObject(unsafe.staticFieldBase(field), unsafe.staticFieldOffset(field), value);
			} catch (Exception e) {
				throw new ReflectiveOperationException(e);
			}
		}

		public static <T> T getField(Object obj, Field field) throws ReflectiveOperationException {
			if (obj == null) {
				return getStaticField(field);
			} else {
				try {
					return (T)unsafe.getObject(obj, unsafe.objectFieldOffset(field));
				} catch (Exception e) {
					throw new ReflectiveOperationException(e);
				}
			}
		}

		public static <T> T getStaticField(Field field) throws ReflectiveOperationException {
			try {
				unsafe.ensureClassInitialized(field.getDeclaringClass());
				return (T)unsafe.getObject(unsafe.staticFieldBase(field), unsafe.staticFieldOffset(field));
			} catch (Exception e) {
				throw new ReflectiveOperationException(e);
			}
		}
	}
}
